"""Tests for `PDBNucleicAcids` package."""

import unittest

import PDBNucleicAcids


class TestPDBNucleicAcids(unittest.TestCase):
    """Tests for `PDBNucleicAcids` package."""

    # def setUp(self):
    #     """Set up test fixtures, if any."""

    # def tearDown(self):
    #     """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""
        assert PDBNucleicAcids.__version__
